#include <cstdlib>
#include <cstring>
#include "SDL.h"
#include "compression.h"
#include "pcximage.h"
#include "fcnc_endian.h"
#include "inifile.h"
#include "shpimage.h"
#include "vfs.h"
#include "logger.h"

//#define DEBUG_PCX


PCXImage::PCXImage(const char* fname, int scaleq) : pcxdata(0), image(0) {

	image				= NULL;
	HeaderError			= false;
	Use16ColorPalette	= true;

	VFile* imgfile;
	this->scaleq = scaleq;
	imgfile = VFS_Open(fname);
	if (imgfile == NULL) {
		logger->error ("%s line %i: Image not found %s\n", __FILE__, __LINE__, fname);
		throw ImageNotFound();
	}
	imgsize = imgfile->fileSize();

	// The image should at least be bigger that the header
	if (imgsize  < 129){
		VFS_Close(imgfile);
		logger->error ("%s line %i: Not an pcx image %s\n", __FILE__, __LINE__, fname);
		HeaderError = true;
		return;
	}

	pcxdata = new Uint8[imgsize];
	imgfile->readByte(pcxdata, imgsize);

	header.Signature	= readbyte(pcxdata,0);
	header.Version		= readbyte(pcxdata,1);
	header.Encoding		= readbyte(pcxdata,2);
	header.BitsPixel	= readbyte(pcxdata,3);
	header.XMin			= readword(pcxdata,4);
	header.YMin			= readword(pcxdata,6);
	header.XMax			= readword(pcxdata,8);
	header.YMax			= readword(pcxdata,10);
	header.HRes			= readword(pcxdata,12);
	header.VRes			= readword(pcxdata,14);

	// The image should at least be bigger that the header
	if (header.Signature !=  10){
		VFS_Close(imgfile);
		logger->error ("%s line %i: Not an pcx image %s\n", __FILE__, __LINE__, fname);
		HeaderError = true;
		return;
	}

	// Read the palette
	Uint32 offset = 16;
	for (int i = 0; i < 16; i++){
		header.palette[i].r = readbyte(pcxdata, offset);
		header.palette[i].g = readbyte(pcxdata, offset+1);
		header.palette[i].b = readbyte(pcxdata, offset+2);
		offset += 3;
	}

	header.Reserved		= readbyte(pcxdata,64);
	header.NumPlanes	= readbyte(pcxdata,65);
	header.BytesLine	= readword(pcxdata,66);
	header.PalType		= readword(pcxdata,68);
	header.HscreenSize	= readword(pcxdata,70);
	header.VscreenSize	= readword(pcxdata,72);

#ifdef DEBUG_PCX
	Uint32 Height	= header.YMax-header.YMin+1;
	Uint32 Width	= header.BytesLine*header.NumPlanes; //Width:= XMax-XMin+1)
	logger->note ("%s line %i: PRINTING pcx header for file: %s\n", __FILE__, __LINE__, fname);
	logger->note ("%s line %i: Signature \t\t= %i\n", __FILE__, __LINE__, header.Signature);
	logger->note ("%s line %i: Version \t\t= %i\n", __FILE__, __LINE__, header.Version);
	logger->note ("%s line %i: Encoding \t\t= %i\n", __FILE__, __LINE__, header.Encoding);
	logger->note ("%s line %i: BitsPixel \t\t= %i\n", __FILE__, __LINE__, header.BitsPixel);
	logger->note ("%s line %i: XMin \t\t\t= %i\n", __FILE__, __LINE__, header.XMin);
	logger->note ("%s line %i: YMin \t\t\t= %i\n", __FILE__, __LINE__, header.YMin);
	logger->note ("%s line %i: XMax \t\t\t= %i\n", __FILE__, __LINE__, header.XMax);
	logger->note ("%s line %i: YMax \t\t\t= %i\n", __FILE__, __LINE__, header.YMax);
	logger->note ("%s line %i: HRes \t\t\t= %i\n", __FILE__, __LINE__, header.HRes);
	logger->note ("%s line %i: VRes \t\t\t= %i\n", __FILE__, __LINE__, header.VRes);
	logger->note ("%s line %i: NumPlanes \t\t= %i\n", __FILE__, __LINE__, header.NumPlanes);
	logger->note ("%s line %i: BytesLine \t\t= %i\n", __FILE__, __LINE__, header.BytesLine);
	logger->note ("%s line %i: PalType \t\t= %i\n", __FILE__, __LINE__, header.PalType);
	logger->note ("%s line %i: HscreenSize \t= %i\n", __FILE__, __LINE__, header.HscreenSize);
	logger->note ("%s line %i: VscreenSize \t= %i\n", __FILE__, __LINE__, header.VscreenSize);

	logger->note ("%s line %i: Image width \t= %i\n", __FILE__, __LINE__, Width);
	logger->note ("%s line %i: Image height \t= %i\n", __FILE__, __LINE__, Height);
	logger->note ("%s line %i: Image size \t\t= %i\n", __FILE__, __LINE__, Height*Width);
#endif

	// Read the vga pallette if present
	readPalette();

    VFS_Close(imgfile);
}

PCXImage::~PCXImage()
{
	if (pcxdata != NULL)
		delete[] pcxdata;

	if (image != NULL)
		SDL_FreeSurface(image);
}

void PCXImage::readPalette()
{
    Uint16 i;
	
	Uint32 offset = imgsize - 769;

#ifdef DEBUG_PCX
	logger->note ("%s line %i: Vga palette indicator = %i (should be 12 if it contains a 256 color palette)\n", __FILE__, __LINE__, readbyte(pcxdata, offset));
#endif

	// This is the vga palette indicator (must be 12 if there is a palette)
	if (readbyte(pcxdata, offset) == 12){
		Use16ColorPalette = false;
		offset++;
		for (i = 0; i < 256; i++) {
			palette[i].r = readbyte(pcxdata, offset);
			palette[i].g = readbyte(pcxdata, offset+1);
			palette[i].b = readbyte(pcxdata, offset+2);
			offset += 3;
		}
	}
}

SDL_Surface* PCXImage::getImage()
{
	if (HeaderError)
		return NULL;

    if (image == NULL) {
        loadImage();
    }
    return image;
}

void PCXImage::loadImage()
{
unsigned char Pixel;
unsigned char Number;
unsigned int y, x;
long Index;
unsigned int offset;
SDL_Surface* imgtmp;
Uint8 *imgdst;


	Uint32 Height	= header.YMax-header.YMin+1;
	Uint32 Width	= header.BytesLine*header.NumPlanes; //header.XMax-header.XMin+1;

	// Image data var ( + Width to be absolutely sure ;) )
	imgdst = new Uint8[Height*Width+Width];

	/* RLE (Run Length Encoding) */
	offset = 128;
	Index = 0;
	y = 0;
	while( y < Height ) {
		x = 0;
		while( x < Width ){
			Pixel = readbyte( pcxdata, offset );
			offset++;
			if( Pixel > 192 ){
				Number = Pixel-192;
				Pixel = readbyte( pcxdata, offset );
				offset++;
				for( unsigned int i=0; i<Number; i++ ){
					imgdst[Index++] = Pixel;
					x++;
				}
			}else{
				imgdst[Index++] = Pixel;
				x++;
			}
		}
		y++;
	}

	//imgtmp = SDL_CreateRGBSurfaceFrom( imgdst, header.HRes, header.VRes, header.BitsPixel, header.BytesLine, 0, 0, 0, 0 );
	imgtmp = SDL_CreateRGBSurfaceFrom( imgdst, header.XMax-header.YMin, header.YMax-header.XMin, header.BitsPixel, header.BytesLine, 0, 0, 0, 0 );
	if (!Use16ColorPalette)
		SDL_SetColors(imgtmp,palette,0,256);
	else
		SDL_SetColors(imgtmp,header.palette,0,16);

	SDL_SetColorKey(imgtmp,SDL_SRCCOLORKEY,0);

	if (scaleq >= 0) {
		image = scaler.scale(imgtmp,scaleq);
		SDL_SetColorKey(image,SDL_SRCCOLORKEY,0);
	} else {
		image = SDL_DisplayFormat(imgtmp);
	}
	SDL_FreeSurface(imgtmp);
	delete[] imgdst;
}

